local moduleName = 'ActionBars'
local idTweaks = LibStub('AceAddon-3.0'):GetAddon('idTweaks')
local module = idTweaks:newModule(moduleName, 'AceConsole-3.0', 'AceEvent-3.0')
local LibActionButton = LibStub('LibActionButton-1.0')

local TL, TC, TR = 'TOPLEFT', 'TOP', 'TOPRIGHT'
local ML, MC, MR = 'LEFT', 'CENTER', 'RIGHT'
local BL, BC, BR = 'BOTTOMLEFT', 'BOTTOM', 'BOTTOMRIGHT'

module.defaults = {
  profile = {
    enabled = true,
    barPadding = 0,
    barMargin = 0,
    buttonPadding = 0,
  }
}

module.options = {
  name = moduleName,
  type = 'group',
  args = {
    enabled = {
      order = 1,
      name = 'Enabled',
      desc = 'WARNING: Reloads the UI!',
      type = 'toggle',
      get = function()
        return module.db.profile.enabled
      end,
      set = function(_, value)
        local MinimapModule = idTweaks:GetModule('Minimap')
        module.db.profile.enabled = value
        module:check()
        MinimapModule.shape.position:check()

        ReloadUI()
      end,
    },

    -- size = {
    --   name = 'Size',
    --   desc = 'Control the size of the frame.',
    --   type = 'group',
    --   args = {
    --     barMargin = {
    --       name = 'Bar Margin',
    --       type = 'range',
    --       min = 0,
    --       max = 50,
    --       step = 1,
    --       bigStep = 1,
    --       get = function()
    --         return module.db.profile.barMargin
    --       end,
    --       set = function(_, value)
    --         module.db.profile.barMargin = value
    --         module:sizeAllBars()
    --         module:positionAllBars()
    --       end
    --     },
    --
    --     barPadding = {
    --       name = 'Bar Padding',
    --       type = 'range',
    --       min = 0,
    --       max = 50,
    --       step = 1,
    --       bigStep = 1,
    --       get = function()
    --         return module.db.profile.barPadding
    --       end,
    --       set = function(_, value)
    --         module.db.profile.barPadding = value
    --         module:sizeAllBars()
    --         module:positionAllBars()
    --       end
    --     },
    --
    --     buttonPadding = {
    --       name = 'Button Padding',
    --       type = 'range',
    --       min = 0,
    --       max = 50,
    --       step = 1,
    --       bigStep = 1,
    --       get = function()
    --         return module.db.profile.buttonPadding
    --       end,
    --       set = function(_, value)
    --         module.db.profile.buttonPadding = value
    --         module:sizeAllBars()
    --         module:positionAllBars()
    --       end
    --     },
    --   },
    -- },
  },
}

function module:initialize()
  module.frames = {}
end

function module:enable()
  module:disableBlizzardBars()
  module:enableActionBars()
  module:enablePetBar()

  module:sizeAllBars()
  module:positionAllBars()
end

function module:disable()
  module:disablePetBar()
  module:disableActionBars()
  module:enableBlizzardBars()
end

function module:registerKeybindings()
  BINDING_HEADER_IDTBAR1 = 'Bar 1'
  BINDING_HEADER_IDTBAR2 = 'Bar 2'
  BINDING_HEADER_IDTBAR3 = 'Bar 3'
  BINDING_HEADER_IDTBAR4 = 'Bar 4'
  BINDING_HEADER_IDTBAR5 = 'Bar 5'
  BINDING_HEADER_IDTPETBAR = 'Pet Bar'
  local x = 1
  for i=1,5 do
    for j=1,12 do
      _G['BINDING_NAME_CLICK idTweaksActionBar' .. i .. 'Button' .. j .. ':' .. x] = j
      x = x + 1
    end
  end
  for j=1,10 do
    _G['BINDING_NAME_CLICK idTweaksActionBarPetButton' .. j .. ':' .. j] = j
  end
end

function module:enableActionBars()
  module.frames.bar1 = module.frames.bar1 or module:createBar('idTweaksActionBar1', 'action', 1, 12)
  module.frames.bar2 = module.frames.bar2 or module:createBar('idTweaksActionBar2', 'action', 13, 24)
  module.frames.bar3 = module.frames.bar3 or module:createBar('idTweaksActionBar3', 'action', 25, 36)
  module.frames.bar4 = module.frames.bar4 or module:createBar('idTweaksActionBar4', 'action', 37, 48)
  module.frames.bar5 = module.frames.bar5 or module:createBar('idTweaksActionBar5', 'action', 49, 60)
end

function module:disableActionBars()
  if module.frames.bar1 then
    module.frames.bar1:ClearAllPoints()
    module.frames.bar1:SetPoint(BR, UIParent, TL, 0, 0)
  end
end

function module:enablePetBar()
  module.frames.petbar = module.frames.petbar or module:createPetBar('idTweaksActionBarPet', 'action', 1, 10)
  module.frames.petbar:Hide()

  local function petEvent(event)
    if InCombatLockdown() then return end

    if PetHasActionBar() then
      if not module.frames.petbar:IsShown() then
        module.frames.petbar:Show()
      end
    else
      module.frames.petbar:Hide()
    end
  end

  module:RegisterEvent('PLAYER_CONTROL_LOST', petEvent)
  module:RegisterEvent('PLAYER_CONTROL_GAINED', petEvent)
  module:RegisterEvent('PLAYER_FARSIGHT_FOCUS_CHANGED', petEvent)
  module:RegisterEvent('UNIT_PET', petEvent)
  module:RegisterEvent('UNIT_FLAGS', petEvent)
  module:RegisterEvent('PET_BAR_UPDATE', petEvent)
  module:RegisterEvent('PET_BAR_UPDATE_COOLDOWN', petEvent)
  module:RegisterEvent('PET_BAR_SHOWGRID', petEvent)
  module:RegisterEvent('PET_BAR_HIDEGRID', petEvent)
  module:RegisterEvent('PET_BAR_UPDATE_USABLE', petEvent)
  module:RegisterEvent('PET_UI_UPDATE', petEvent)
  module:RegisterEvent('PLAYER_TARGET_CHANGED', petEvent)
  module:RegisterEvent('PLAYER_MOUNT_DISPLAY_CHANGED', petEvent)
end

function module:disablePetBar()
  if not module.frames.petbar then
    return
  end

  module:UnregisterEvent('PLAYER_CONTROL_LOST')
  module:UnregisterEvent('PLAYER_CONTROL_GAINED')
  module:UnregisterEvent('PLAYER_FARSIGHT_FOCUS_CHANGED')
  module:UnregisterEvent('UNIT_PET')
  module:UnregisterEvent('UNIT_FLAGS')
  module:UnregisterEvent('PET_BAR_UPDATE')
  module:UnregisterEvent('PET_BAR_UPDATE_COOLDOWN')
  module:UnregisterEvent('PET_BAR_SHOWGRID')
  module:UnregisterEvent('PET_BAR_HIDEGRID')
  module:UnregisterEvent('PET_BAR_UPDATE_USABLE')
  module:UnregisterEvent('PET_UI_UPDATE')
  module:UnregisterEvent('PLAYER_TARGET_CHANGED')
  module:UnregisterEvent('PLAYER_MOUNT_DISPLAY_CHANGED')

  PetActionBarFrame:EnableMouse('true')
  SlidingActionBarTexture0:SetAlpha(1)
  SlidingActionBarTexture1:SetAlpha(1)

  for i = module.frames.petbar.firstId, module.frames.petbar.lastId do
    local button = _G['PetActionButton' .. i]
    button:ClearAllPoints()
    button:SetParent(PetActionBarFrame)
    button:SetPoint(BR, PetActionBarFrame, BL, 65 + ((i - 1) * 38), 0)
  end

  module.frames.petbar:Hide()
end

function module:enableBlizzardBars()
  MainMenuBar:SetScript('OnShow', nil)
  MainMenuBar:Show()

  MultiBarLeft:SetScript('OnShow', nil)
  MultiBarLeft:Show()

  MultiBarRight:SetScript('OnShow', nil)
  MultiBarRight:Show()
end

function module:disableBlizzardBars()
  MainMenuBar:SetScript('OnShow', MainMenuBar.Hide)
  MainMenuBar:Hide()

  MultiBarLeft:SetScript('OnShow', MultiBarLeft.Hide)
  MultiBarLeft:Hide()

  MultiBarRight:SetScript('OnShow', MultiBarRight.Hide)
  MultiBarRight:Hide()
end

function module:createButton(i, buttonName, frameHeader)
  local button = LibActionButton:CreateButton(i, buttonName, frameHeader, {
    outOfRangeColoring = 'button',
    tooltip = 'enabled',
    showGrid = false,
    colors = {
      range = { 0.8, 0.1, 0.1 },
      mana = { 0.5, 0.5, 1.0 }
    },
    hideElements = {
      macro = false,
      hotkey = false,
      equipped = false,
    },
    keyBoundTarget = false,
    clickOnDown = false,
    flyoutDirection = 'UP',
  })

  local icon = _G[buttonName .. 'Icon']
  -- local texture = (_G[buttonName .. 'NormalTexture2'] or _G[buttonName .. 'NormalTexture'])
  -- local hotkey = _G[buttonName .. 'HotKey']
  -- local macrotext = _G[buttonName .. 'Name']

  icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

  return button
end

function module:createBar(barName, barType, firstId, lastId)
  local frameHeader = CreateFrame('Frame', barName .. 'Header', UIParent, 'SecureHandlerStateTemplate')
  local bar = CreateFrame('Frame', barName, UIParent)

  bar.header = frameHeader
  bar.barName = barName
  bar.barType = barType
  bar.buttonName = barName .. 'Button'
  bar.firstId = firstId
  bar.lastId = lastId

  bar.background = bar:CreateTexture(barName .. 'Background', 'BACKGROUND')
  bar.background:SetAllPoints(bar)
  bar.background:SetColorTexture(0, 0, 0, 0.5)

  local j = 1
  for i = bar.firstId, bar.lastId do
    bar['button' .. j] = module:createButton(i, bar.buttonName .. j, bar.header)
    j = j + 1
  end

  module:sizeBar(bar)
  module:setBarActions(bar, bar.barType, firstId, lastId)

  RegisterStateDriver(bar.header, 'page', '0')
  bar.header:SetAttribute('_onstate-page', [[
    module:SetAttribute('state', newstate)
    control:ChildUpdate('state', newstate)
  ]])

  return bar
end

function module:createPetBar(barName, barType, firstId, lastId)
  local frameHeader = CreateFrame('Frame', barName .. 'Header', UIParent, 'SecureHandlerStateTemplate')
  local bar = CreateFrame('Frame', barName, UIParent)

  bar.header = frameHeader
  bar.barName = barName
  bar.barType = barType
  bar.buttonName = barName .. 'Button'
  bar.firstId = firstId
  bar.lastId = lastId

  bar.background = bar:CreateTexture(barName .. 'Background', 'BACKGROUND')
  bar.background:SetAllPoints(bar)
  bar.background:SetColorTexture(0, 0, 0, 0.5)

  local j = 1
  for i = bar.firstId, bar.lastId do
    local button = _G['PetActionButton' .. i]
    local buttonIcon = _G['PetActionButton' .. i .. 'Icon']
    local buttonBorder = _G['PetActionButton' .. i .. 'NormalTexture2']

    button:ClearAllPoints()
    button:SetParent(bar)
    button:SetWidth(idTweaksActionBar1Button1:GetWidth())
    button:SetHeight(idTweaksActionBar1Button1:GetHeight())

    buttonIcon:SetTexCoord(0.08,0.92,0.08,0.92)
    buttonBorder:Hide()

    if button:HasScript('OnUpdate') then
      button:HookScript('OnUpdate', function()
        if module.frames.petbar then
          buttonBorder:Hide()
        end
      end)
    end

    bar['button' .. j] = button

    j = j + 1
  end

  PetActionBarFrame:EnableMouse('false')
  SlidingActionBarTexture0:SetAlpha(0)
  SlidingActionBarTexture1:SetAlpha(0)

  module:sizeBar(bar)

  return bar
end

function module:setBarActions(bar, barType, firstId, lastId)
  for i = 1, lastId - firstId + 1 do
    bar['button'..i]:SetState(0, barType, firstId - 1 + i)
  end
end

function module:sizeBar(bar)
  local barPadding = module.db.profile.barPadding
  local buttonPadding = module.db.profile.buttonPadding

  local j = 1
  for i = bar.firstId, bar.lastId do
    local button = bar['button' .. j]

    if j == 1 then
      button:SetPoint(ML, bar, ML, barPadding, 0)
    else
      button:SetPoint(ML, bar['button' .. j - 1], MR, buttonPadding, 0)
    end

    j = j + 1
  end

  local buttonAmount = bar.lastId + 1 - bar.firstId
  local buttonWidth = bar.button1:GetWidth()
  local buttonHeight = bar.button1:GetHeight()

  local width = buttonWidth * buttonAmount + buttonPadding * (buttonAmount - 1) + barPadding * 2
  local height = buttonHeight + barPadding * 2

  bar:SetWidth(width)
  bar:SetHeight(height)
end

function module:sizeAllBars()
  for b = 1, 5 do
    local bar = module.frames['bar'..b]

    module:sizeBar(bar)
  end
end

function module:getTotalSize()
  local width = module.frames.bar1:GetWidth()
  local height = 5 * module.frames.bar1:GetHeight()

  return width, height
end

function module:positionAllBars()
  local barMargin = module.db.profile.barMargin

  module.frames.bar1:ClearAllPoints()

  if idTweaks.modules.ExperienceBar.db.profile.enabled then
    module.frames.bar1:SetPoint(BC, idTweaksExperienceBar, TC, 0, barMargin)
  else
    module.frames.bar1:SetPoint(BC, UIParent, BC, 0, barMargin)
  end

  module.frames.bar2:SetPoint(BC, module.frames.bar1, TC, 0, barMargin)
  module.frames.bar3:SetPoint(BC, module.frames.bar2, TC, 0, barMargin)
  module.frames.bar4:SetPoint(BC, module.frames.bar3, TC, 0, barMargin)
  module.frames.bar5:SetPoint(BC, module.frames.bar4, TC, 0, barMargin)
  module.frames.petbar:SetPoint(BC, module.frames.bar5, TC, 0, barMargin)
end
