local moduleName = 'BlizzardUI'
local idTweaks = LibStub('AceAddon-3.0'):GetAddon('idTweaks')
local module = idTweaks:newModule(moduleName, 'AceConsole-3.0', 'AceHook-3.0')

module.defaults = {
  profile = {
    enabled = true,

    buffFrame = {
      hide = true,
    },

    castingBarFrame = {
      hide = true,
    },

    errorsFrame = {
      abilityIsNotReadyYet = true,
      outOfRange = true,
      spellIsNotReadyYet = true,
      targetNeedsToBeInFrontOfYou = true,
      targetTooClose = true,
      thereIsNothingToAttack = true,
      youAreFacingTheWrongWay = true,
      youAreTooFarAway = true,
      youCantDoThatYet = true,
      youHaveNoTarget = true,
    }
  }
}

module.options = {
  name = moduleName,
  type = 'group',
  args = {
    enabled = {
      order = 1,
      name = 'Enabled',
      desc = 'Enables / Disables the Module',
      type = 'toggle',
      get = function()
        return module.db.profile.enabled
      end,
      set = function(_, value)
        module.db.profile.enabled = value
        module:check()
      end,
    },

    buffFrame = {
      name = 'Buffs',
      type = 'group',
      disabled = function()
        return not module.db.profile.enabled
      end,
      args = {
        hide = {
          name = 'Hide',
          type = 'toggle',
          get = function()
            return module.db.profile.buffFrame.hide
          end,
          set = function(_, value)
            module.db.profile.buffFrame.hide = value
            module.buffFrame:check()
          end
        },
      }
    },

    castingBarFrame = {
      name = 'Casting Bar',
      type = 'group',
      disabled = function()
        return not module.db.profile.enabled
      end,
      args = {
        hide = {
          name = 'Hide',
          type = 'toggle',
          get = function()
            return module.db.profile.castingBarFrame.hide
          end,
          set = function(_, value)
            module.db.profile.castingBarFrame.hide = value
            module.castingBarFrame:check()
          end
        },
      }
    },

    errorsFrame = {
      name = 'Errors',
      type = 'group',
      disabled = function()
        return not module.db.profile.enabled
      end,
      args = {
        abilityIsNotReadyYet = {
          name = 'Ability is not ready yet.',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.abilityIsNotReadyYet
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.abilityIsNotReadyYet = value
            module.errorsFrame:check()
          end
        },

        outOfRange = {
          name = 'Out of range.',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.outOfRange
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.outOfRange = value
            module.errorsFrame:check()
          end
        },

        spellIsNotReadyYet = {
          name = 'Spell is not ready yet.',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.spellIsNotReadyYet
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.spellIsNotReadyYet = value
            module.errorsFrame:check()
          end
        },

        targetNeedsToBeInFrontOfYou = {
          name = 'Target needs to be in front of you.',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.targetNeedsToBeInFrontOfYou
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.targetNeedsToBeInFrontOfYou = value
          end
        },

        targetTooClose = {
          name = 'Target too close',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.targetTooClose
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.targetTooClose = value
          end
        },

        thereIsNothingToAttack = {
          name = 'There is nothing to attack.',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.thereIsNothingToAttack
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.thereIsNothingToAttack = value
          end
        },

        youAreFacingTheWrongWay = {
          name = 'You are facing the wrong way!',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.youAreFacingTheWrongWay
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.youAreFacingTheWrongWay = value
          end
        },

        youAreTooFarAway = {
          name = 'You are too far away!',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.youAreTooFarAway
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.youAreTooFarAway = value
          end
        },

        youCantDoThatYet = {
          name = 'You can\'t do that yet',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.youCantDoThatYet
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.youCantDoThatYet = value
          end
        },

        youHaveNoTarget = {
          name = 'You have no target.',
          type = 'toggle',
          get = function()
            return module.db.profile.errorsFrame.youHaveNoTarget
          end,
          set = function(_, value)
            module.db.profile.errorsFrame.youHaveNoTarget = value
          end
        },
      }
    }
  }
}

function module:enable()
  module.buffFrame:check()
  module.castingBarFrame:check()
  -- module.errorsFrame:check()
end

function module:disable()
  module.buffFrame:disable()
  module.castingBarFrame:disable()
  -- module.errorsFrame:disable()
end

module.buffFrame = {}

function module.buffFrame:check()
  if module.db.profile.buffFrame.hide then
    module.buffFrame:enable()
  else
    module.buffFrame:disable()
  end
end

function module.buffFrame:enable()
  module:SecureHook(BuffFrame, 'Show', function(frame)
    if (module.db.profile.buffFrame.hide) then
      BuffFrame:Hide()
    end
  end)
  if (module.db.profile.buffFrame.hide) then
    BuffFrame:Hide()
  end
end

function module.buffFrame:disable()
  module:Unhook(BuffFrame, 'Show')

  BuffFrame:Show()
end

module.castingBarFrame = {}

function module.castingBarFrame:check()
  if module.db.profile.castingBarFrame.hide then
    module.castingBarFrame:enable()
  else
    module.castingBarFrame:disable()
  end
end

function module.castingBarFrame:enable()
  module:SecureHook(CastingBarFrame, 'Show', function(frame)
    if (module.db.profile.castingBarFrame.hide) then
      CastingBarFrame:Hide()
    end
  end)
end

function module.castingBarFrame:disable()
  module:Unhook(CastingBarFrame, 'Show')
end

module.errorsFrame = {}

module:RawHook(UIErrorsFrame, 'AddMessage', function(frame, message, red, green, blue, messageId, holdTime)
  if not module.db.profile.enabled then
    module.hooks[frame].AddMessage(frame, message, red, green, blue, messageId, holdTime)
    return
  end

  for k, v in pairs(module.db.profile.errorsFrame) do
    local option = module.options.args.errorsFrame.args[k]

    if v and option and message == option.name then
      return
    end
  end

  module.hooks[frame].AddMessage(frame, message, red, green, blue, messageId, holdTime)
end, true)

-- function module.errorsFrame:check()
--
-- function module.errorsFrame:enable()
-- end
--
-- function module.errorsFrame:disable()
-- end
