local moduleName = 'Mail'
local idTweaks = LibStub('AceAddon-3.0'):GetAddon('idTweaks')
local module = idTweaks:newModule(moduleName, 'AceConsole-3.0', 'AceEvent-3.0')

module.defaults = {
  profile = {
    enabled = true,
    autoSendMail = true
  }
}

module.options = {
  name = moduleName,
  type = 'group',
  args = {
    enabled = {
      order = 1,
      name = 'Enabled',
      desc = 'Enables / Disables the Module',
      type = 'toggle',
      get = function()
        return module.db.profile.enabled
      end,
      set = function(_, value)
        module.db.profile.enabled = value
        module:check()
      end,
    },

    autoSendMail = {
      name = 'Auto Send',
      desc = 'When the mail has a Recipient and all item slots are filled, send it.',
      type = 'toggle',
      disabled = function()
        return not module.db.profile.enabled
      end,
      get = function()
        return module.db.profile.autoSendMail
      end,
      set = function(_, value)
        module.db.profile.autoSendMail = value
      end
    }
  },
}

function module:enable()
  module:RegisterEvent('MAIL_SEND_INFO_UPDATE')
end

function module:disable()
  module:UnregisterEvent('MAIL_SEND_INFO_UPDATE')
end

function module:getFreeMailSlots()
  local freeSlots = 0

  for slot = 1, 12 do
    itemName, _, _, _ = GetSendMailItem(slot)

    if not itemName then
      freeSlots = freeSlots + 1
    end
  end

  return freeSlots
end

function module:mailIsFull()
  return module:getFreeMailSlots() == 0
end

function module:getMailRecipient()
  return SendMailNameEditBox:GetText()
end

function module:setMailRecipient(recipient)
  SendMailNameEditBox:SetText(recipient)
end

function module:mailHasRecipient()
  return module:getMailRecipient() ~= ''
end

function module:getMailSubject()
  return SendMailSubjectEditBox:GetText()
end

function module:mailHasSubject()
  return module:getMailSubject() ~= ''
end

function module:getMailBody()
  return SendMailScrollChildFrame:GetText()
end

function module:canSendMail()
  return module:mailHasRecipient() and module:mailHasSubject() and module:mailIsFull()
end

function module:sendMail()
  local recipient = module:getMailRecipient()
  local subject = module:getMailSubject()
  -- local body = module:getMailBody()

  SendMail(recipient, subject, '')

  module:setMailRecipient(recipient)
end

function module:MAIL_SEND_INFO_UPDATE()
  if module.db.profile.autoSendMail == false then
    return
  end

  if module:canSendMail() then
    module:sendMail()
  end
end
