local moduleName = 'UnitFrames'
local idTweaks = LibStub('AceAddon-3.0'):GetAddon('idTweaks')
local module = idTweaks:newModule(moduleName, 'AceConsole-3.0', 'AceHook-3.0')
local LibSharedMedia = LibStub('LibSharedMedia-3.0')

local _, ns = ...
local oUF = ns.oUF

local TL, TC, TR = 'TOPLEFT', 'TOP', 'TOPRIGHT'
local ML, MC, MR = 'LEFT', 'CENTER', 'RIGHT'
local BL, BC, BR = 'BOTTOMLEFT', 'BOTTOM', 'BOTTOMRIGHT'

local STATUS_BAR_TEXTURE = 'Interface/TargetingFrame/UI-StatusBar'

local sharedMediaStatusBars = LibSharedMedia:HashTable('statusbar')
local sharedMediaStatusBarsReversed = idTweaks.table.reverse(LibSharedMedia:HashTable('statusbar'))
local sharedMediaBackgrounds = LibSharedMedia:HashTable('background')
local sharedMediaBackgroundsReversed = idTweaks.table.reverse(LibSharedMedia:HashTable('background'))

local default = {
  frame = {
    -- TODO: Move top level size to here
    size = {
    },

    -- TODO: Implement.
    positioning = {
      from = 'CENTER',
      target = 'UIParent',
      to = 'CENTER',
      x = 0,
      y = 0,
    },

    -- TODO: Implement.
    backgroundTexture = '',

    -- TODO: Implement.
    borderTexture = '',
    borderInsets = {
      top = 0,
      left = 0,
      right = 0,
      bottom = 0,
    },
  },

  size = {
    width = 600,
    height = 40,
    padding = 5,
    margin = 5,
  },

  healthBar = {
    texture = 'Interface\\TargetingFrame\\UI-StatusBar',

    -- TODO: Rename to backgroundTexture
    background = 'Interface\\DialogFrame\\UI-DialogBox-Background-Dark',

    -- TODO: Implement.
    borderTexture = '',

    -- TODO: Implement.
    borderInsets = {
      top = 0,
      left = 0,
      right = 0,
      bottom = 0,
    },
  },
  powerBar = {
    texture = 'Interface\\TargetingFrame\\UI-StatusBar',

    -- TODO: Rename to backgroundTexture
    background = 'Interface\\DialogFrame\\UI-DialogBox-Background-Dark',

    -- TODO: Implement.
    borderTexture = '',

    -- TODO: Implement.
    borderInsets = {
      top = 0,
      left = 0,
      right = 0,
      bottom = 0,
    },
  },
  additionalPowerBar = {
    texture = 'Interface\\TargetingFrame\\UI-StatusBar',

    -- TODO: Rename to backgroundTexture
    background = 'Interface\\DialogFrame\\UI-DialogBox-Background-Dark',

    -- TODO: Implement.
    borderTexture = '',

    -- TODO: Implement.
    borderInsets = {
      top = 0,
      left = 0,
      right = 0,
      bottom = 0,
    },
  },
  text = {
    -- TODO: Implement.
    name = {
      padding = 5,
      font = '',
      fontColor = { r=255, g=255, b=255 },
      tag = '',
    },

    -- TODO: Implement.
    health = {
      padding = 5,
      font = '',
      fontColor = { r=255, g=255, b=255 },
      tag = '',
    },

    -- TODO: Implement.
    power = {
      padding = 5,
      font = '',
      fontColor = { r=255, g=255, b=255 },
      tag = '',
    },

    -- TODO: Implement.
    classPower = {
      padding = 5,
      font = '',
      fontColor = { r=255, g=255, b=255 },
      tag = '',
    },
  },
}

module.defaults = {
  profile = {
    enabled = true,

    default = default,

    pet = idTweaks.table.merge(default, {
      defaults = true
    }),

    player = idTweaks.table.merge(default, {
      defaults = true
    }),

    target = idTweaks.table.merge(default, {
      defaults = true
    }),

    targettarget = idTweaks.table.merge(default, {
      defaults = true
    }),

    party = idTweaks.table.merge(default, {
      defaults = true
    }),
  },
}

module.options = {
  name = 'UnitFrames',
  type = 'group',
  args = {
    enabled = {
      name = 'Enabled',
      type = 'toggle',
      get = function()
        return module.db.profile.default.enabled
      end,
      set = function(_, value)
        module.db.profile.default.enabled = value
      end
    },
    default = module:getUnitOptions('Default', 'default'),
    pet = module:getUnitOptions('Pet', 'pet'),
    player = module:getUnitOptions('Player', 'player'),
    target = module:getUnitOptions('Target', 'target'),
    targettarget = module:getUnitOptions('Target of Target', 'targettarget'),
    party = module:getUnitOptions('Party', 'party'),
  }
}

-- Utility functions.
function module:getDefaultOrUnitSetter(fn)
  return function(self, unit, ...)
    if unit == 'default' then
      for name, frame in pairs(module.frames) do
        if module.db.profile[name].defaults then
          fn(name, ...)
        end
      end
    else
      fn(unit, ...)
    end
  end
end

function module:shouldFrameHaveAdditionalPower(unit)
  local _, _, classIndex = UnitClass(unit)

  if classIndex == 1 then
  elseif classIndex == 2 then
  end

  --   Death Knight:  PowerBar and Runes (text)
  --   Demon Hunter:  PowerBar
  --   Druid:         PowerBar and AdditionalPowerBar (Moonkin)
  -- * Mage           PowerBar and Arcane Charges (text)
  -- * Monk           PowerBar and Chi (text)
  -- * Paladin        PowerBar Holy Power (text)
  --   Priest         Mana ???
  -- * Rogue          PowerBar Combo Points (text)
  --   Shaman         PowerBar and AdditionalPowerBar ???
  -- * Warlock        PowerBar and Soul Shards (text)
  --   Warrior        PowerBar
end

function module:getDBForUnit(unit)
  local dbUnit = module.db.profile[unit]
  local dbDefault = module.db.profile.default

  return dbUnit.defaults and dbDefault or dbUnit
end

function module:getUnitOptions(unitName, unit)
  local output = {
    name = unitName,
    type = 'group',
    args = {
      enabled = {
        order = 1,
        name = 'Enabled',
        desc = 'Enables / Disables the Module',
        type = 'toggle',
        get = function()
          return module.db.profile.enabled
        end,
        set = function(_, value)
          module.db.profile.enabled = value
          module:check()
        end,
      },

      size = {
        name = 'Size',
        desc = 'Control the size of the frame.',
        type = 'group',
        disabled = function()
          return unit ~= 'default' and module.db.profile[unit].defaults
        end,
        args = {
          width = {
            name = 'Width',
            type = 'range',
            min = 100,
            max = 1000,
            step = 1,
            bigStep = 10,
            get = function()
              return module.db.profile[unit].size.width
            end,
            set = function(_, value)
              module.db.profile[unit].size.width = value
              module.size:positionElements(unit)
            end
          },
          height = {
            name = 'Height',
            type = 'range',
            min = 5,
            max = 100,
            step = 1,
            bigStep = 1,
            get = function()
              return module.db.profile[unit].size.height
            end,
            set = function(_, value)
              module.db.profile[unit].size.height = value
              module.size:positionElements(unit)
            end
          },
          padding = {
            name = 'Padding',
            desc = 'Spacing between elements in the frame.',
            type = 'range',
            min = 0,
            max = 50,
            step = 1,
            bigStep = 1,
            get = function()
              return module.db.profile[unit].size.padding
            end,
            set = function(_, value)
              module.db.profile[unit].size.padding = value
              module.size:positionElements(unit)
            end
          },
          margin = {
            name = 'Margin',
            desc = 'Spacing between this frame and the next.',
            type = 'range',
            min = 0,
            max = 50,
            step = 1,
            bigStep = 1,
            get = function()
              return module.db.profile[unit].size.margin
            end,
            set = function(_, value)
              module.db.profile[unit].size.margin = value
              module.size:positionFrame(unit)
            end
          },
        },
      },
      positioning = {
        name = 'Positioning',
        desc = 'Control the positioning of the frame.',
        disabled = function() return unit ~= 'default' and module.db.profile[unit].defaults end,
        type = 'group',
        args = {
        },
      },
      healthBar = {
        name = 'Health Bar',
        disabled = function()
          return unit ~= 'default' and module.db.profile[unit].defaults
        end,
        type = 'group',
        args = {
          texture = {
            name = 'Texture',
            type = 'select',
            dialogControl = 'LSM30_Statusbar',
            values = LibSharedMedia:HashTable('statusbar'),
            get = function()
              return sharedMediaStatusBarsReversed[module.db.profile[unit].healthBar.texture]
            end,
            set = function(_, value)
              module.db.profile[unit].healthBar.texture = sharedMediaStatusBars[value]
              module.healthBar:setTexture(unit)
            end
          },
          background = {
            name = 'Texture',
            type = 'select',
            dialogControl = 'LSM30_Background',
            values = LibSharedMedia:HashTable('background'),
            get = function()
              return sharedMediaBackgroundsReversed[module.db.profile[unit].healthBar.background]
            end,
            set = function(_, value)
              module.db.profile[unit].healthBar.background = sharedMediaBackgrounds[value]
              module.healthBar:setBackground(unit)
            end
          },
        },
      },
      powerBar = {
        name = 'Power Bar',
        disabled = function()
          return unit ~= 'default' and module.db.profile[unit].defaults
        end,
        type = 'group',
        args = {
          width = {
            name = 'Texture',
            type = 'select',
            dialogControl = 'LSM30_Statusbar',
            values = LibSharedMedia:HashTable('statusbar'),
            get = function()
              return sharedMediaStatusBarsReversed[module.db.profile[unit].powerBar.texture]
            end,
            set = function(_, value)
              module.db.profile[unit].powerBar.texture = sharedMediaStatusBars[value]
              module.powerBar:setTexture(unit)
            end
          },
          background = {
            name = 'Texture',
            type = 'select',
            dialogControl = 'LSM30_Background',
            values = LibSharedMedia:HashTable('background'),
            get = function()
              return sharedMediaBackgroundsReversed[module.db.profile[unit].powerBar.background]
            end,
            set = function(_, value)
              module.db.profile[unit].powerBar.background = sharedMediaBackgrounds[value]
              module.powerBar:setBackground(unit)
            end
          },
        },
      },
      additionalPowerBar = {
        name = 'Additional Power Bar',
        disabled = function()
          return unit ~= 'default' and module.db.profile[unit].defaults
        end,
        type = 'group',
        args = {
          width = {
            name = 'Texture',
            type = 'select',
            dialogControl = 'LSM30_Statusbar',
            values = LibSharedMedia:HashTable('statusbar'),
            get = function()
              return sharedMediaStatusBarsReversed[module.db.profile[unit].additionalPowerBar.texture]
            end,
            set = function(_, value)
              module.db.profile[unit].additionalPowerBar.texture = sharedMediaStatusBars[value]
              module.additionalPowerBar:setTexture(unit)
            end
          },
          background = {
            name = 'Texture',
            type = 'select',
            dialogControl = 'LSM30_Background',
            values = LibSharedMedia:HashTable('background'),
            get = function()
              return sharedMediaBackgroundsReversed[module.db.profile[unit].additionalPowerBar.background]
            end,
            set = function(_, value)
              module.db.profile[unit].additionalPowerBar.background = sharedMediaBackgrounds[value]
              module.additionalPowerBar:setBackground(unit)
            end
          },
        },
      },
      text = {
        name = 'Text',
        disabled = function()
          return unit ~= 'default' and module.db.profile[unit].defaults
        end,
        type = 'group',
        args = {
        },
      },
    },
  }

  if unit == 'default' then
  else
    output.args.defaults = {
      name = 'Defaults',
      type = 'toggle',
      get = function()
        return module.db.profile[unit].defaults
      end,
      set = function(_, value)
        module.db.profile[unit].defaults = value
      end
    }
  end

  return output
end

-- Keeps the instances of the frames.
module.frames = {}

-- Size functions.
module.size = {}

module.size.positionElements = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  --[[
    total     = 120
    padding   =   5
    healthBar =  60
    padding   =   5
    powerBar  =  20
    padding   =   5
    aPowerBar =  20
    padding   =   5

    120 (total) - 25 (padding) = 95

    95 / 2 = 47.5

    47.5 / 3

    total     = 145
  ]]

  frame:SetWidth(db.size.width)
  frame:SetHeight(db.size.height)

  local widthMinusPadding = db.size.width - db.size.padding * 2
  local heightMinusPadding = db.size.height - db.size.padding * 3 -- 2 statusbars

  frame.Health:SetPoint(TL, frame, TL, db.size.padding, -db.size.padding)
  frame.Health:SetWidth(widthMinusPadding)
  frame.Health:SetHeight(heightMinusPadding / 2)

  frame.Power:SetPoint(TC, frame.Health, BC, 0, -db.size.padding)
  frame.Power:SetWidth(widthMinusPadding)
  frame.Power:SetHeight(heightMinusPadding / 2)

  -- frame.AdditionalPower:SetPoint(TC, frame.Power, BC, 0, -db.size.padding)
  -- frame.AdditionalPower:SetWidth(widthMinusPadding)
  -- frame.AdditionalPower:SetHeight(heightMinusPadding / 2 / 2)

  frame.Buffs.size = db.size.width / 8
  frame.Buffs.initialAnchor = BL
  frame.Buffs['growth-x'] = 'RIGHT'
  frame.Buffs['growth-y'] = 'UP'
  frame.Buffs:SetSize(8 * frame.Buffs.size, 2 * frame.Buffs.size)
  frame.Buffs:SetPoint(BL, frame, TL, 0, db.size.padding)

  frame.Debuffs.size = db.size.width / 8
  frame.Debuffs.initialAnchor = TL
  frame.Debuffs['growth-x'] = 'RIGHT'
  frame.Debuffs['growth-y'] = 'DOWN'
  frame.Debuffs:SetSize(8 * frame.Debuffs.size, 2 * frame.Debuffs.size)
  frame.Debuffs:SetPoint(TL, frame, BL, 0, -db.size.padding)
end)

module.size.positionFrame = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  if unit == 'player' then
    frame:SetPoint(MR, UIParent, MC, -db.size.margin / 2, -200)
  elseif unit == 'target' then
    frame:SetPoint(ML, UIParent, MC, db.size.margin / 2, -200)
  elseif unit == 'targettarget' then
    frame:SetPoint(ML, module.frames.target, MR, db.size.margin, 0)
  elseif unit == 'pet' then
    print('pet!', MR, ML, -db.size.margin, 0, module.frames.player)
    frame:SetPoint(MR, module.frames.player, ML, -db.size.margin, 0)

  -- elseif unit == 'party' then
  --   frame:SetPoint(TC, module.frames.pet, BC, 0, -margin)
  end
end)

module.healthBar = {}

module.healthBar.setTexture = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  frame.Health:SetStatusBarTexture(db.healthBar.texture)

end)

module.healthBar.setBackground = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  frame.Health.bg:SetTexture(db.additionalPowerBar.background)
end)

module.powerBar = {}

module.powerBar.setTexture = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  frame.Power:SetStatusBarTexture(db.powerBar.texture)
end)

module.powerBar.setBackground = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  frame.Power.bg:SetTexture(db.additionalPowerBar.background)
end)

module.additionalPowerBar = {}

module.additionalPowerBar.setTexture = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  frame.AdditionalPower:SetStatusBarTexture(db.additionalPowerBar.texture)
end)

module.additionalPowerBar.setBackground = module:getDefaultOrUnitSetter(function(unit, frame)
  frame = frame or module.frames[unit]
  local db = module:getDBForUnit(unit)

  frame.AdditionalPower.bg:SetTexture(db.additionalPowerBar.background)
end)

-- Positioning functions.
module.positioning = {}

-- oUF Layout functions.
module.layouts = {
  unit = {},
}

function module.layouts.default(self, unit)
  -- idTweaks:Print('module.layouts.default', unit, self)

  -- TODO: Make the Name Text be the Spell Name when Casting
  -- TODO: Threat
  -- TODO: Icons

  local db = module:getDBForUnit(unit)

  -- Background
  module.background = module:CreateTexture(nil, 'BACKGROUND')
  module.background:SetAllPoints(self)
  module.background:SetColorTexture(0, 0, 0, 0.5)

  -- Health Bar
  local Health = CreateFrame('StatusBar', nil, self)
  Health.colorReaction = true
  Health.colorClass = true
  module.Health = Health

  -- Health Bar Background
  local HealthBG = Health:CreateTexture(nil, 'BORDER')
  HealthBG:SetAllPoints(Health)
  Health.bg = HealthBG

  -- Power Bar
  local Power = CreateFrame('StatusBar', nil, self)
  Power.frequentUpdates = true
  Power.displayAltPower = true
  Power.colorPower = true
  module.Power = Power

  -- Power Bar Background
  local PowerBG = Health:CreateTexture(nil, 'BORDER')
  PowerBG:SetAllPoints(Power)
  Power.bg = PowerBG

  -- Additional Power Bar
  -- local AdditionalPower = CreateFrame('StatusBar', nil, self)
  -- AdditionalPower.colorPower = true
  -- module.AdditionalPower = AdditionalPower

  -- Additional Power Bar Background
  -- local AdditionalPowerBG = Health:CreateTexture(nil, 'BORDER')
  -- AdditionalPowerBG:SetAllPoints(AdditionalPower)
  -- AdditionalPower.bg = AdditionalPowerBG

  -- Buffs
  local Buffs = CreateFrame('Frame', nil, self)
  Buffs.num = 16
  module.Buffs = Buffs

  -- Debuffs
  local Debuffs = CreateFrame('Frame', nil, self)
  Debuffs.num = 16
  module.Debuffs = Debuffs

  -- Tags
  local textPadding = 5
  local NameText = module.Health:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
  NameText:SetFont(NameText:GetFont(), 9)
  NameText:SetTextColor(1, 1, 1, 1)
  NameText:SetPoint(ML, module.Health, ML, textPadding, 0)
  NameText:SetHeight(9)
  module:Tag(NameText, '[name][ >leader][ >status][ >threatcolor][threat]|r')
  module.NameText = NameText

  local HealthText = module.Health:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
  HealthText:SetFont(HealthText:GetFont(), 9)
  HealthText:SetTextColor(1, 1, 1, 1)
  HealthText:SetPoint(MR, module.Health, MR, -textPadding, 0)
  HealthText:SetHeight(9)
  module:Tag(HealthText, '[curhp]')
  module.HealthText = HealthText

  oUF.Tags.Methods['runes'] = function(unit, realUnit)
    if unit ~= 'player' then return end
    local amount = 0
    for i=1,6 do
      local _, _, ready = GetRuneCooldown(i)
      if ready == true then
        amount = amount + 1
      end
    end
    return amount
  end

  -- * Death Knight:  PowerBar and [runes]
  -- * Demon Hunter:  PowerBar
  --   Druid:         PowerBar and [cpoints] (Cat) and AdditionalPowerBar (Mana for Moonkin)
  -- * Mage           PowerBar and [arcanecharges]
  -- * Monk           PowerBar and [chi]
  -- * Paladin        PowerBar and [holypower]
  --   Priest         Mana ???
  -- * Rogue          PowerBar and [cpoints]
  --   Shaman         PowerBar and AdditionalPowerBar (Mana)
  -- * Warlock        PowerBar and [soulshards]
  -- * Warrior        PowerBar

  local ClassPowerText = module.Power:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
  ClassPowerText:SetFont(ClassPowerText:GetFont(), 9)
  ClassPowerText:SetTextColor(1, 1, 1, 1)
  ClassPowerText:SetPoint(ML, module.Power, ML, textPadding, 0)
  ClassPowerText:SetHeight(9)
  module:Tag(ClassPowerText, '[runes][arcanecharges][chi][holypower][soulshards][cpoints]')
  module.ClassPowerText = ClassPowerText

  local PowerText = module.Power:CreateFontString(nil, 'OVERLAY', 'GameFontNormal')
  PowerText:SetFont(PowerText:GetFont(), 9)
  PowerText:SetTextColor(1, 1, 1, 1)
  PowerText:SetPoint(MR, module.Power, MR, -textPadding, 0)
  PowerText:SetHeight(9)
  module:Tag(PowerText, '[curpp]')
  module.PowerText = PowerText

  module.size:positionElements(unit, self)
  module.size:positionFrame(unit, self)
  module.healthBar:setTexture(unit, self)
  module.healthBar:setBackground(unit, self)
  module.powerBar:setTexture(unit, self)
  module.powerBar:setBackground(unit, self)
  -- module.additionalPowerBar:setTexture(unit, self)
  -- module.additionalPowerBar:setBackground(unit, self)

  if(module.layouts.unit[unit]) then
    return module.layouts.unit[unit](self)
  end
end

function module.layouts.factory(self)
  -- idTweaks:Print('module.layouts.factory')

  module:SetActiveStyle('idTweaks')

  module.frames.player = module:Spawn('player')
  module.frames.pet = module:Spawn('pet')
  module.frames.target = module:Spawn('target')
  module.frames.targettarget = module:Spawn('targettarget')

  -- TODO: Enable Party
  -- oUF:SpawnHeader(overrideName, overrideTemplate, visibility, attributes ...)
  -- module.frames.party = module:SpawnHeader(nil, nil, 'raid,party,solo',
  --   -- http://wowprogramming.com/docs/secure_template/Group_Headers
  --   -- Set header attributes
  --   'showParty', true,
  --   'showPlayer', true,
  --   'yOffset', -0
  -- )
end
